#
# Original Author
# Steve Spicklemire
# Silicon Prairie Ventures Inc. (SPVI)
#
# Significant Contributions by Shane Hathaway, Digital Creations Inc.
#
# Portions (Copyright (C) 2001, Mikael BERTHE, Alcove)
# 
# Disclaimer
#   
#   THIS SOFTWARE IS PROVIDED BY SPVI ``AS IS'' AND ANY EXPRESSED OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED.  IN NO EVENT SHALL SPVI OR ITS CONTRIBUTORS BE LIABLE FOR
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
# 
# 

from ZCVSMixin import ZCVSMixin, do_break
from Globals import HTMLFile, MessageDialog
import OFS
import os

pathsep = os.sep
pathseptwice = pathsep + pathsep

class CVSFolder(ZCVSMixin, OFS.SimpleItem.Item):
    id = 'CVS'
    meta_type = 'CVS Folder'
    
    def _containerObjectIds(self):
        return filter(lambda id: id != 'CVS', self.aq_parent.objectIds())

    def _containerHasID(self, id):
        """ Check to see if container has this ID """
        if self.isInZClass:
            parent = self.aq_parent
            return parent.getClassAttr(id, parent) is not parent
        else:
            return id in self._containerObjectIds()

    def _container_p_jar(self):
        return self.aq_parent._p_jar

    def _containerDelObject(self, id):
        if id != 'CVS':  # For sanity
            self.aq_parent._delObject(id)

    def _containerSetObject(self, id, value):
        if id != 'CVS':  # For sanity
            self.aq_parent._setObject(id, value)

    def _containerGetattr(self, id):
        if id != 'CVS':  # For sanity
            if self.isInZClass:
                attr = self.aq_parent[id]
            else:
                attr = getattr(self.aq_parent, id)
            return attr
        else:
            raise AttributeError, id
    
    def _setId(self, id):
        if id != self.id:
            raise MessageDialog(
                title='Invalid Id',
                message='Cannot change the id of a CVSFolder',
                action ='./manage_main',)

    def __init__(self, fsroot, fsrepos, addfsdir, usepassfile, simpleInterface,
                 passfile, default_gargs, isInZClass=0, useplainfiles=0):
        self.fsroot = fsroot + pathsep
        self.fsrepos = fsrepos
        self.addfsdir = addfsdir
        self.usepassfile = usepassfile
        self.simpleInterface = simpleInterface
        self.passfile=passfile
        self.default_gargs = default_gargs
        self.isInZClass = isInZClass
        self.useplainfiles = useplainfiles

        import re

        self.fsrepos = re.sub(pathseptwice + '+', pathsep, self.fsrepos)
        self.fsroot = re.sub(pathseptwice + '+', pathsep, self.fsroot)
        if self.fsrepos and self.fsrepos[0] == pathsep:
            self.fsrepos = self.fsrepos[1:]
            
    def branchToCVS(self, root=1):
        """
        Convert a branch to CVS
        """
        failures = []
        newzcvsf = []

        zobj = self._containerObjectIds()

        for id in zobj:
            thisObj = self._containerGetattr(id).aq_base
            if hasattr(thisObj, 'isAnObjectManager') and thisObj.isAnObjectManager and thisObj.meta_type != 'ZCatalog':
                if ('CVS' in thisObj.objectIds()):
                    # There is already a CVS object
                    failures.append('%s' %id)
                else:
                    # Let's create the CVS object
                    ob = CVSFolder( self.getFSRoot(),
                        self.fsrepos + os.sep + id,
                        self.addfsdir, self.usepassfile,
                        self.passfile, self.default_gargs, self.isInZClass, self.useplainfiles)

                    if self.addfsdir:
                        ob.active = 1
                        ob.manage_addfsdirCVS(self.default_gargs)

                    if self.isInZClass:
                        ob.id = ob.id + '_ZClass'

                    thisObj._setObject(ob.id, ob)
                    newzcvsf.append('%s' %id)

                if hasattr(thisObj.CVS, 'branchToCVS'):
                    succ, fail = thisObj.CVS.branchToCVS(0)
                else:
                    succ, fail = [], [id]
                    
                newzcvsf = newzcvsf + succ
                failures = failures + fail

        if root:
            return MessageDialog(title='Branch -> CVS',
                     message= 'New ZCVS folders: %s<BR>' %newzcvsf + \
                     '<BR>List of skipped folders: %s' % failures,
                     action='cvsStatusForm')
        return newzcvsf,failures


manage_addCVSFolderForm = HTMLFile('addCVSFolder', globals())

def manage_addCVSFolder(self, fsroot, fsrepos, addfsdir=0, usepassfile=0, 
                        simpleInterface=0, passfile='', default_gargs='',
                        isInZClass=0, useplainfiles=0, REQUEST=None, URL1='Youre_lost'):

    '''Adds a CVS folder to the current folder.
    '''

    #if hasattr(self.aq_parent, 'CVS'):
    if 'CVS' in self.aq_parent.objectIds():
        return MessageDialog(
                 title  ='Item Exists',
                 message='This object already contains a CVS Folder',
                 action=URL1+'/manage_main')

    # do_break()

    if not fsrepos:
        return MessageDialog(
                 title  ='Need a Value!',
                 message="File System Repository Can't be Empty",
                 action=URL1+'/manage_main')

    ob = CVSFolder(fsroot, fsrepos, addfsdir, usepassfile, simpleInterface,
                   passfile, default_gargs, isInZClass)

    if addfsdir:
        ob.active = 1
        result = ob.manage_addfsdirCVS(default_gargs)
    else:
        result = 'no new directory added'

    if isInZClass:
        ob.id = ob.id + '_ZClass'

    self._setObject(ob.id, ob)

    return MessageDialog(title='OK',
                         message = ' Added CVS Folder... <pre> add result: ' \
                                    + result + '</pre>', action='manage_main')

