#
# Original Author
# Steve Spicklemire
# Silicon Prairie Ventures Inc. (SPVI)
#
# Significant Contributions by Shane Hathaway, Digital Creations Inc.
#
# Portions (Copyright (C) 2001, Mikael BERTHE, Alcove)
# 
# Disclaimer
#   
#   THIS SOFTWARE IS PROVIDED BY SPVI ``AS IS'' AND ANY EXPRESSED OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED.  IN NO EVENT SHALL SPVI OR ITS CONTRIBUTORS BE LIABLE FOR
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
# 
# 

import os
import sys
import select
import tempfile
import string

try:
    import zLOG

except:
    class foo:
        pass

    def LOG( subsystem, severity, short ):
        if severity > -500:
            print subsystem, severity, short

    zLOG = foo()
    zLOG.LOG = LOG
    zLOG.BLATHER = 0
    zLOG.INFO = 1


from threading import Thread
class Reader(Thread):
    "Thread for reading a file in the background"
    def __init__(self, fd):
        Thread.__init__(self)
        self.fd = fd

    def run(self):
        self.result = self.fd.read()
        self.fd.close()


class ShellCommandProcessor:

    active = 0
    usepassfile = 0
    useplainfiles = 0

    def getUsePlainFiles(self, PARENTS=None):

        result = None

        if PARENTS:
            for index in range(len(PARENTS)):
                try:
                    # get the highest possible 'root'
                    result = PARENTS[index].__dict__.get('CVS',None).useplainfiles
                except AttributeError:
                    pass
            else:
                #do_break()
                try:
                    result = self.aq_parent.aq_parent.CVS.getUsePlainFiles()
                except AttributeError:
                    pass

            if result is not None:
                return result

        return self.useplainfiles

    def processStreams(self, command):

        if not self.active:
            if self._v_debug:
                zLOG.LOG('ZCVSMixin', zLOG.BLATHER,
                         "Not processing streams... self.active = 0")
                
            # no commands are executed unless this is true...
            return  0, '', ''

        if self._v_debug:
            self.DBGwrite("Running command '%s'... " % command)

        if self.usepassfile:
            os.environ['CVS_PASSFILE'] = self.passfile

        if self.useplainfiles:

            #
            # Zope doesn't like forking... so let's just be dumb and use files
            # rather than pipes..
            #
            # <mgedmin@delfi.lt>: Actually the previous fork+pipe code was a
            # little buggy -- nonblocking I/O is fundamentally incompatible with
            # Python's <fileobject>.read() due to buffering performed by C
            # library's fread().  I rewrote the code to use blocking reads in
            # two different threads, let's hope it works better.  Another
            # posibility would be to use os.read() with nonblocking I/O and
            # avoid any buffering, but it's more difficult and I do not see any
            # advantages that would offset the complexity.
            #

            stdoutFname = tempfile.mktemp()
            stderrFname = tempfile.mktemp()

            if type(command) != type(''):
                command = " ".join(command)     # FIXME: shell quoting?
            iresult = self.doCommand(command, stdoutFname, stderrFname)

            zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'checking result .. "%i"' % iresult)

            resultStd = open(stdoutFname).read()
            resultErr = open(stderrFname).read()

            os.remove(stdoutFname)
            os.remove(stderrFname)

            zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'Got result.. "%s"' % resultStd)
            zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'Got err result.. "%s"' % resultErr)

            return iresult, resultStd, resultErr

        # Create a subprocess
        from popen2 import Popen3
        childInst = Popen3(command, 1, -1)
        child_stdout, child_stdin = childInst.fromchild, childInst.tochild
        child_stderr = childInst.childerr
        child_stdin.close()

        # Create a thread for reading stderr in the background
        reader = Reader(child_stderr)
        reader.start()

        # Read stdout in the foreground
        resultStd = child_stdout.read()
        child_stdout.close()

        # Wait for the child to finish
        reader.join()
        resultErr = reader.result
        resultVal = childInst.wait()

        return resultVal, resultStd, resultErr

    def processCommands(self, command):
        return self.processStreams(command)[0]

    def DBGwrite(self, str):
        if self._v_debug:
            dfdbg = open("/tmp/zopeCVS.log", "a+")
            dfdbg.write(str)
            dfdbg.close()

    def doCommand(self, command, stdoutFname, stderrFname):

        cmd = 'sh -v -c "%s" >%s 2>%s' % (command, stdoutFname, stderrFname)

        iresult = os.system(cmd)
        return iresult

if __name__=='__main__':

    import pdb
    pdb.set_trace()
    
    x = ShellCommandProcessor()
    x.active = 1
    x._v_debug = 1
    x.useplainfiles = 1

    print x.processStreams('ls -lga')
