# 
# ZCVS Mixin Base Class
#
# Original Author
# Steve Spicklemire
# Silicon Prairie Ventures Inc. (SPVI)
#
# Significant Contributions by Shane Hathaway, Digital Creations Inc.
#
# Portions (Copyright (C) 2001, Mikael BERTHE, Alcove)
#
# Yet More Significant Contributions by Itai Tavor
#
# Disclaimer
#   
#   THIS SOFTWARE IS PROVIDED BY SPVI ``AS IS'' AND ANY EXPRESSED OR
#   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED.  IN NO EVENT SHALL SPVI OR ITS CONTRIBUTORS BE LIABLE FOR
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
# 
# 

__doc__ = "ZCVS Mixin Base Class"
__version__ = "0.2.4"

try:
    from ZCVSExIm import doExport
except:
    def doExport(a, b):
        pass

import types
from ShellCommandProcessor import ShellCommandProcessor
import StringIO
import traceback

try:

    from OFS import XMLExportImport
    customImporters={
        XMLExportImport.magic: XMLExportImport.importXML,
    }
except:
    customImporters = None

try:
    import Globals, OFS, AccessControl, zLOG
    from Acquisition import Implicit
    from Globals import Persistent, HTMLFile, MessageDialog, ImageFile
    from DocumentTemplate import html_quote

    def do_break():
        import pdb
        pdb.set_trace()

    class Dummy (Persistent,
                 Implicit,
                 AccessControl.Role.RoleManager,
                 ShellCommandProcessor):
        pass

except:

    class foo:
        pass

    class HTMLFile:

        def __init__(self, *args):
            pass

    class ImageFile:
        def __init__(self, *args):
            pass

    def LOG( subsystem, severity, short ):
        if severity > -500:
            print subsystem, severity, short

    zLOG = foo()
    zLOG.LOG = LOG
    zLOG.BLATHER = 0
    zLOG.INFO = 1

    class Dummy(ShellCommandProcessor):
        """ for testing outside Zope... """

        def objectIds(self):
            return ['hehe','nuther']

        hehe = 'foo'
        nuther = 'blah'

        def _getOb(self, id):
            return None


import os
import string
import time
import tempfile

statDict = {
    'Up-to-date':'OK',
    'Locally Modified':'LM',
    'Locally Added':'LA',
    'Locally Removed':'LR',
    'Needs Checkout':'NC',
    'Needs Patch':'NP',
    'Unknown':'UN',
    }

class TimeTracker:

    """
    Keep track of times... we have:

    fsmtime  (mtime from filesystem...) and
    bmtime  (bobbase modification time for 'real' modification)

    When an object is pulled from cvs it gets a bobobase
    modification time of 'now'. CVS also has a concept of
    modification time. Assuming nobody monkeys with files
    on the filesystem other than cvs.. we can 'fake' real
    modification times by deliberately keeping track of
    fsmtime and bmtime.

    """

    fsmtime = 0.0
    bmtime = 0.0

    def ZopeChanged(self, curr_z_bmtime):
        """ put in 5 seconds leeway.. """
        return curr_z_bmtime > self.bmtime + 5

    def FSChanged(self, curr_fs_mtime):
        """ put in 5 seconds leeway.. """
        return curr_fs_mtime > self.fsmtime + 5

    def ZopeChangedTime(self, curr_z_bmtime):

        if self.ZopeChanged():
            return curr_z_bmtime
        else:
            return self.fsmtime

    def FSChangedTime(self, curr_fs_mtime):
        if self.FSChanged(curr_fs_mtime):
            return curr_fs_mtime
        else:
            return self.fsmtime


DEFAULTPASSFILE='/var/zope/.cvspass'

class ZCVSMixin(Dummy):

    """ZCVS Mixin"""

    meta_type = "ZCVS Mixin Base Class"
    fsroot = '/tmp'
    fsrepos = ''
    addfsdir = 1
    isInZClass = 0          # special tricks for ZClasses...
    getFSRootFromParent = 1 # try to get FSRoot from parent class
    default_gargs = '-z6'
    allow_gargs = 0
    allowEmptyCommit = 0
    simpleInterface = 0
    trackerDict = None
    passfile = DEFAULTPASSFILE
    usepassfile = 1
    cleanupXML = 0
    # we don't do these (yet?)
    badTypes = ('ZCVS Folder', 'Folder', 'CVS Folder', 'User Folder')
    excludeTypes = []
    _v_debug = 0

    __ac_permissions__ = (

        ('Manage CVS User',
            ['manage_checkIn', 'manage_checkOut',
             'cvsUserForm',
             'manage_objFromFS', 'manage_objToFS',
             'manage_cvsAdmin'],
            ['Manager']),
        ('Manage CVS Controls',
            ['manage_fcheckOut', 'cvsStatusForm',
             'manage_addCVS', 'manage_objFromFS', 'manage_objToFS',
             'manage_remove', 'manage_removeAll'],
            ['Manager']),
        ('Manage CVS Controls Admin',
            ['cvsAdminForm', 'manage_addfsdirCVS', 'manage_editCVS',
             'branchToCVS', 'cvsActionForm', 'setAdminOptions'],
            ['Manager']),
        )

    def manage_options(self):
        options = self._manage_options
        if self.simpleInterface:
            return (options[0], options[2], options[3])
        else:
            return (options[1], options[2], options[3])

    _manage_options = (
        {'label':'CVS User', 'action':'cvsUserForm'},
        {'label':'CVS Status', 'action':'cvsStatusForm'},
        {'label':'CVS Actions', 'action':'cvsActionForm'},
        {'label':'CVS Admin', 'action':'cvsAdminForm'},
        )

    cvsAdminForm = HTMLFile('cvsAdminForm', globals())
    cvsActionForm = HTMLFile('cvsActionForm', globals())
    cvsStatusForm = HTMLFile('cvsStatusForm', globals())
    cvsUserForm = HTMLFile('cvsUserForm', globals())

    cvs_folder_icon = ImageFile('cvs_folder_icon.gif', globals())

    same_icon = ImageFile('same.gif', globals())
    diff_icon = ImageFile('diff.gif', globals())

    left_blue_icon = ImageFile('left_blue.gif', globals())
    left_green_icon = ImageFile('left_green.gif', globals())
    left_red_icon = ImageFile('left_red.gif', globals())

    right_blue_icon = ImageFile('right_blue.gif', globals())
    right_green_icon = ImageFile('right_green.gif', globals())
    right_red_icon = ImageFile('right_red.gif', globals())

    def excludedTypes(self):
        return list(self.badTypes) + self.excludeTypes

    def _containerObjectIds(self):
        return self.objectIds()

    def _containerHasID(self, id):
        return id in self.objectIds()

    def _container_p_jar(self):
        return self._p_jar

    def _containerDelObject(self, id):
        self._delObject(id)

    def _containerSetObject(self, id, value):
        self._setObject(id, value)

    def _containerGetattr(self, id):
        return getattr(self, id)

    def checkTracker(self):
        if self.trackerDict is None:
            self.trackerDict = {}

    def reloadAdmin(self):
        """ dynamic reload of admin forms... for development.. """
        self.cvsAdminForm = HTMLFile('cvsAdminForm', globals())
        self.cvsStatusForm = HTMLFile('cvsStatusForm', globals())
        return "OK!"

    def getFSRoot(self, PARENTS=None):

        result = None

        if self.getFSRootFromParent:
            if PARENTS:
                for index in range(len(PARENTS)):
                    try:
                        # get the highest possible 'root'
                        result = PARENTS[index].__dict__.get('CVS',None).fsroot
                    except AttributeError:
                        pass
            else:
                #do_break()
                try:
                    result = self.aq_parent.aq_parent.CVS.getFSRoot()
                except AttributeError:
                    pass

            if result is not None:
                return result

        if hasattr(self, 'fsdir'): # backward compatibility.....
            return os.path.split(self.fsdir)[0]
        else:
            return self.fsroot

    def getFSRepos(self):
        if hasattr(self, 'fsdir'):
            return '/' + os.path.split(self.fsdir)[1]
        else:
            return self.fsrepos

    def getMetaType(self, obj):
        """ get a meta type.... call if necessary """
        result = ''

        if hasattr(obj, 'meta_type'):
            mt = obj.meta_type
            if hasattr(mt, 'im_func'):
                mt = mt()
            result = mt

        return result

    def getNewFSRepos(self, PARENTS=None):
        """
        try to determine the new repository name looking into Parents...
        """

        base = ''
        isZClass = 0

        mt = self.getMetaType(PARENTS[1])

        if mt == 'Z Class':
            #
            # Ack... special case... fix up for ZClass
            #
            foundIndex = -1
            isZClass = 1

            for obj in PARENTS:
                if obj.__class__.__name__ == 'Product':
                    break
                elif obj.__class__.__name__ == 'ZClass':

                    if base:
                        base = os.sep + base

                    base = obj.id + base
        else:
            base = PARENTS[1].id

        for index in range(len(PARENTS)):
            try:
                base = PARENTS[index].__dict__.get('CVS',None).fsrepos + \
                       os.sep + base
                break
            except AttributeError:
                pass

        return base, isZClass

    def getFSDir(self):
        if hasattr(self, 'fsdir'):
            return self.fsdir
        else:
            return self.getFSRoot() + self.getFSRepos()

    def checkList(self, obj_id):
        """
        check for list type... convert if necessary.
        """

        if type(obj_id) == type(''):
            obj_id = string.split(obj_id)
        elif obj_id is None:
            obj_id = []

        return obj_id

    def isCVSContainer(self, obj):
        """
        Check if object can/should contain CVSFolder instances
        """
        return getattr(obj, 'isPrincipiaFolderish', 0) and \
            obj.meta_type != 'ZCatalog'

    def idsToFileNames(self, statusDict, relNames=None):
        fnames = []
        for obj_id in statusDict.keys():
            if statusDict.has_key(obj_id) and \
               statusDict[obj_id].has_key('ext'):
                ext = statusDict[obj_id]['ext']
            else:
                ext = 'xml'

            if relNames:
                fnames.append("%s.%s" % (obj_id,ext))
            else:
                fnames.append(os.path.join(self.getFSDir(),
                              ("%s.%s" % (obj_id, ext))))

        return fnames

    def idToFileName(self, obj_id, ext, relName=None):
        if relName:
            return "%s.%s" % (obj_id, ext)
        else:
            return os.path.join(self.getFSDir(), "%s.%s" % (obj_id, ext))

    def getTimeTracker(self, obj_id):
        """
        get recursively an object time tracker...
        """
        if string.find(obj_id, '/') != -1:
            # Recursive call
            folder, newobj_id = string.split(obj_id, '/', 1)
            try:
                thisObj = self._containerGetattr(folder)
            except:
                return None
            if hasattr(thisObj, 'isAnObjectManager') and \
               thisObj.isAnObjectManager:
                return thisObj.CVS.getTimeTracker(newobj_id)
            else:
                raise 'fileToObject', "%s is not an object manager" %folder
        # make sure we have a trackerDict...
        self.checkTracker()
        td = self.trackerDict
        return td.get(string.strip(obj_id), TimeTracker())

    def setTimeTracker(self, obj_id, newtracker):
        """
        set (recursively) an object time tracker...
        """
        if string.find(obj_id, '/') != -1:
            # Recursive call
            folder, newobj_id = string.split(obj_id, '/', 1)
            thisObj = self._containerGetattr(folder)
            if hasattr(thisObj, 'isAnObjectManager') and \
               thisObj.isAnObjectManager:
                return thisObj.CVS.getTimeTracker(newobj_id)
            else:
                raise 'fileToObject', "%s is not an object manager" %folder
        # make sure we have a trackerDict...
        self.checkTracker()
        td = self.trackerDict
        td[string.strip(obj_id)] = newtracker
        self.trackerDict = td
        return

    def setAdminOptions(self, allow_gargs, default_gargs, allowEmptyCommit,
                        simpleInterface):
        """
        Recursively set admin options to CVS Folders
        """

        zobj = self._containerObjectIds()

        for id in zobj:
            thisObj = self._containerGetattr(id)
            if self.isCVSContainer(thisObj):
                if hasattr(thisObj.aq_base, 'CVS'):
                    thisObj.CVS.allow_gargs = allow_gargs
                    thisObj.CVS.allowEmptyCommit = allowEmptyCommit
                    thisObj.CVS.default_gargs = default_gargs
                    thisObj.CVS.simpleInterface = simpleInterface
                    thisObj.CVS.setAdminOptions(allow_gargs, default_gargs,
                                allowEmptyCommit, simpleInterface)

    def getListObjs(self, recursive, location = ''):
        """
        A "ls -R" for Zope objects...
        """
        res = {}

        zobj = self._containerObjectIds()

        for id in zobj:
            prep = location + id;

            if not res.has_key(prep):
                thisObj = self._containerGetattr(id)
                if hasattr(thisObj,'aq_base') and hasattr(thisObj.aq_base, 'isAnObjectManager') and thisObj.aq_base.isAnObjectManager:
                    if recursive and hasattr(thisObj,'CVS') and thisObj.CVS.meta_type == 'CVS Folder':
                        nres = thisObj.CVS.getListObjs(recursive, prep + '/')
                        for i in nres.keys():
                            res[i] = nres[i]
                else:
                    res[prep] = {'zname':id, 'location':location,
                                 'zobj':thisObj}
        return res

    def parseStatusList(self, resultStd, recursive=0, secondary=None):
        """ parse stdout from 'cvs status'... use any clues
            we can find to figure out what's what. Results
            are returned in a dict of dicts that has object_ids
            as keys and all the info gleaned from cvs in
            a sub-dictionary tied to that key.
        """

        results = {}
        currObject = {}

        #do_break()

        if secondary:
            # we don't care about the file system.. this is a secondary
            fnames = []
            listObjs = {}
        else:
            fnames = os.listdir(os.curdir)
            listObjs = self.getListObjs(recursive)

            if 'CVS' not in fnames:
                #
                # we are not in a cvs directory.. punt...
                #
                return results


        if self._v_debug:
            zLOG.LOG('ZCVSMixin', zLOG.INFO , 'listObjs..\n' + `listObjs`)
            zLOG.LOG('ZCVSMixin', zLOG.INFO , 'fnames..\n' + `fnames`)
            zLOG.LOG('ZCVSMixin', zLOG.INFO ,
                     'result from cvsStatus...\n' + resultStd)

        # For safety: set a bad default value for 'location'
        # Thus, if the location is not found, we'll not destroy another object
        currObject['location'] = '?/'

        for line in string.split(resultStd,'\n'):
            if string.find(line, '==========') != -1:
                if currObject.has_key('zname'):
                    results[currObject['location'] + currObject['zname']] = \
                            currObject
                    currObject = {'location':'?/'}
            elif string.find(line, 'Status:') != -1:
                status = string.strip(string.split(line, 'Status:')[1])
                currObject['status'] = statDict.get(status, status)
                fname = string.strip(string.split(
                            string.split(line, 'File:')[1],'Status:')[0])
                if string.find(fname,'no file') == -1:
                    currObject['fname'] = fname
                    currObject['zname'] = string.join(
                                          string.split(fname,'.')[:-1],'.')
                    currObject['ext'] = string.split(fname,'.')[-1]

            elif string.find(line, 'Working revision') != -1:
                currObject['wRev'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Repository revision') != -1:
                rRev = string.strip(string.split(string.split(line,':')[1])[0])
                currObject['rRev'] = rRev
                if rRev == 'No':
                    rRev = string.strip(string.split(line,':')[1])
                    currObject['rRev'] = rRev
                else:
                    fname2 = string.split(line, self.getFSRepos(),1)[1];
                    if not currObject.has_key('fname'):
                        fname = string.strip(string.join(
                                string.split(string.split(line,':')[1])[1:]))
                        fname = string.join(string.split(
                                os.path.split(fname2)[-1],',')[:-1],',')
                        currObject['fname'] = fname
                        currObject['zname'] = string.join(
                                string.split(fname,'.')[:-1],'.')
                        currObject['ext'] = string.split(fname,'.')[-1]
                    # Get the folder ('location')
                    fname2 = string.split(fname2, fname)[0] # XXX
                    if fname2[0] == '/':
                        fname2 = fname2[1:]
                    currObject['location'] = fname2

            elif string.find(line, 'Sticky Tag') != -1:
                currObject['sTag'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Sticky Date') != -1:
                currObject['sDate'] = string.strip(string.split(line,':')[1])
            elif string.find(line, 'Sticky Options') != -1:
                currObject['sOptions'] = string.strip(string.split(line,':')[1])

        if currObject.has_key('zname'):
            results[currObject['location'] + currObject['zname']] = \
                    currObject

        for fname in fnames:
            theSplit = string.split(fname,'.')
            zobj_id = string.join(theSplit[:-1],'.')
            ext = theSplit[-1]
            if self._v_debug:
                zLOG.LOG('ZCVSMixin', zLOG.INFO,
                         'looking at %s, %s:%s:%s' % (fname, `theSplit`,
                         zobj_id, ext))

            if not results.has_key(zobj_id) and (len(theSplit)>1) and \
               (ext in ('xml', 'zexp')):
                currObject = {}
                currObject['zname'] = zobj_id
                currObject['status'] = 'OR'
                currObject['fname'] = fname
                currObject['ext'] = ext
                currObject['format_opt'] = ext
                results[zobj_id] = currObject

        for id in listObjs.keys():
            if not results.has_key(id):
                results[id] = listObjs[id]
            elif not results[id].has_key('location') and \
               listObjs[id].has_key('location'):
                results[id]['location'] = listObjs[id]['location']


        for k in results.keys():

            if (k[0] == '?'):
                # 'location' has not been found
                raise 'DEBUG', 'Error: obj="%s"' %k
            tracker = self.getTimeTracker(k)
            results[k]['tracker'] = tracker

            objetTrouve = 0

            # get the object
            if (k in listObjs.keys()):
                results[k]['zobj'] = theObj = listObjs[k]['zobj']
                objetTrouve = 1

            if objetTrouve:

                # see if the object is changed....

                if hasattr(theObj, 'meta_type'):
                    results[k]['meta_type'] = theObj.meta_type

                if hasattr(theObj, 'bobobase_modification_time'):
                    btime = theObj.bobobase_modification_time

                    if hasattr(btime, 'im_func'):
                        btime = btime()

                    results[k]['btime'] = btime = btime.timeTime()
                    results[k]['znewerf'] = tracker.ZopeChanged( btime )

                # check for extension options...

                if not results[k].has_key('ext'):
                    if (hasattr(theObj, 'isPrincipiaFolderish') and
                       theObj.isPrincipiaFolderish):

                        results[k]['format_opt'] = 'zexp'
                    else:
                        results[k]['format_opt'] = 'xml'
                else:
                    if results[k].has_key('location') and \
                       results[k].has_key('fname'):
                        ffpath = results[k]['location'] + results[k]['fname']
                        if os.path.exists(ffpath):
                            mtime = os.path.getmtime(ffpath)
                            results[k]['mtime'] = mtime

                            if tracker.FSChanged(mtime):
                                results[k]['fnewerz'] = 1
                        else:
                            results[k]['znewerf'] = 1
                            results[k]['fnewerz'] = 0
            else:
                results[k]['zobj'] = None

            if results[k].has_key('status'):
                if (results[k]['status'] in ('LM', 'LA')):
                    results[k]['fnewerc'] = 1
                if (results[k]['status'] in ('NU', 'NP')):
                    results[k]['cnewerf'] = 1
            else:
                results[k]['status'] = 'UN'


        for k in results.keys():
            if results[k].has_key('meta_type') and \
               (results[k]['meta_type'] in self.excludedTypes()):
                del results[k]
            elif results[k].has_key('fname') and results[k].has_key('location') and \
               os.access(results[k]['location']+results[k]['fname'],os.F_OK):
                results[k]['fileexists'] = 1
            else:
                results[k]['fileexists'] = 0

        return results

    def parseUpdateList(self, resultStd):
        """ parse stdout from cvs update... """

        results = {}
        lines = string.split(resultStd, '\n')
        for line in lines:
            tokens = string.split(line)
            if len(tokens)>1:
                status = tokens[0]
                if status in ['U', 'A', 'R', 'M', 'C', '?']:
                    fname = tokens[1]
                    results[fname] = status

        return results

    def cvsStatusList(self, zrec = 1):
        """
        Return a dictionary of 'status' settings for objects in CVS compared to
        files on the filesystem. Most other functions take dictionaries like
        this to decide what to do....
        """

        # move to CVS directory...
        oldDir = os.getcwd()
        try:
            os.chdir(self.getFSDir())
        except:
            return (0, {}, 'No such dir')

        d_gargs = self.stringToList(self.default_gargs)

        # check the update
        command = ['cvs'] + d_gargs + ['-n', 'update']
        if not zrec: command = command + ['-l']
        resultVal, resultStd, resultErr = self.processStreams(command)

        # get the update list.. don't really update...
        updateList = self.parseUpdateList( resultStd )

        # check the status
        command = ['cvs'] + d_gargs + ['status']
        if not zrec: command = command + ['-l']
        resultVal, resultStd, resultErr = self.processStreams(command)

        result = self.parseStatusList(resultStd, zrec)

        statAgainList = []
        for fname in updateList.keys():
            obj_id = string.join(string.split(fname,'.')[:-1],'.')
            if not result.has_key(obj_id):
                statAgainList.append(fname)

        if statAgainList:
            command = ['cvs'] + d_gargs + ['status'] + statAgainList
            if not zrec: command = command + ['-l']
            resultVal, resultStd, resultErr = self.processStreams(command)

            secondaryResult = self.parseStatusList(resultStd, zrec,
                              secondary=1)

            for k in secondaryResult.keys():
                result[k] = secondaryResult[k]

        # (MiKael) ...to display in alphabetical order
        # cvsStatusForm.dtml uses this field:
        for k in result.keys():
            result[k]['fullPath'] = k

        os.chdir(oldDir)

        return resultVal, result, resultErr

    def manage_editCVS(self, fsdir=None, fsroot=None, fsrepos=None,
                       addfsdir=0, allow_gargs=0, allowEmptyCommit=0,
                       usepassfile=0, active=0, simpleInterface=0,
                       passfile=DEFAULTPASSFILE, default_gargs='',
                       excludeTypes=None, useplainfiles=0,
                       getFSRootFromParent=0, cleanupXML=0, REQUEST=None):
        """
        Check the properties of the cvs object...
        """

        if hasattr(self, 'fsdir'):
            # backward compatible for old instances...
            self.fsdir = fsdir
        else:
            self.fsroot = fsroot + '/'
            self.fsrepos = fsrepos

            import re
            self.fsroot = re.sub('//+', '/', self.fsroot)

        self.addfsdir = addfsdir
        self.allow_gargs = allow_gargs
        self.allowEmptyCommit = allowEmptyCommit
        self.simpleInterface = simpleInterface
        self.usepassfile = usepassfile
        self.passfile=passfile
        self.default_gargs = default_gargs
        self.active = active
        self.getFSRootFromParent = getFSRootFromParent
        self.cleanupXML = cleanupXML
        self.useplainfiles=useplainfiles

        if excludeTypes:
            self.excludeTypes = excludeTypes

        self.setAdminOptions(allow_gargs, default_gargs, allowEmptyCommit,
                             simpleInterface)

        message='CVS Folder params edited.'
        return self.cvsAdminForm(self, REQUEST, manage_tabs_message=message)

    def manage_checkOut(self, obj_ids=None, g_args='', cvs_args='', isrec=1,
            REQUEST=None):
        """
        check out the latest version...

        1) move the latest changes in zope to the filesystem
           (use the btime date/tracker to decide if changes have been made)

        2) cvs update.. with args...

        3) load updated files into Zope..

        """

        # move to CVS directory....
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            return MessageDialog(title='Check Out',
                       message='No object selected',
                       action=REQUEST.form['FormPage'])

        saveDict = {}

        for obj_id in obj_id_list:
            if statDict[obj_id].has_key('znewerf') \
               and statDict[obj_id]['znewerf']:
                saveDict[obj_id] = statDict[obj_id]

        # save to FS
        self.objectsToFiles(saveDict)

        rv, statDict, ev = self.cvsStatusList(isrec)

        updateList = []
        updateDict = {}

        for obj_id in obj_id_list:
            curStat = statDict[obj_id].get('status','na')
            if curStat == 'NP' or curStat == 'NC':
                updateList.append( statDict[obj_id]['location'] +
                    statDict[obj_id]['fname'] )
                updateDict[obj_id] = statDict[obj_id]
        if len(updateList) == 0:
            return MessageDialog(title='Check Out',
                     message='No object to check-out!',
                     action=REQUEST.form['FormPage'])

        command = ['cvs'] + d_gargs + ['update'] + \
                  self.stringToList(cvs_args) + updateList

        # do the update
        iresult, resultStd, resultErr = self.processStreams(command)

        result = self.formatResult(command, iresult, resultStd, resultErr)
        result = result + '<br>\n' + \
                 formatStatus(self.filesToObjects(updateDict), "Imports")

        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % result

        os.chdir(oldDir)

        return MessageDialog(title='Check Out Returned',
                 message='Command:<PRE>%s</PRE><BR>Update result:<BR>%s' \
                 % (srj(command), result),
                    action=REQUEST.form['FormPage'])


    def manage_fcheckOut(self, obj_ids=None, g_args='', cvs_args='', isrec=1,
            REQUEST=None):
        """
        Force CheckOut
        """

        # move to CVS directory....
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            return MessageDialog(title='Check Out',
                       message='No object selected',
                       action=REQUEST.form['FormPage'])

        result = '';

        rv, statDict, ev = self.cvsStatusList(isrec)

        updateDict = {}
        cvsFailures = []
        listCommands = {}

        for obj_id in obj_id_list:
            curStat = statDict[obj_id].get('status','na')
            realName = statDict[obj_id]['location'] + statDict[obj_id]['fname']

            command = ['cvs'] + d_gargs + ['update', '-p'] + \
                      self.stringToList(cvs_args) + [realName]

            # do the update
            iresult, resultStd, resultErr = self.processStreams(command)
            cmd = ''
            for word in command:
                cmd = cmd + word + ' '
            listCommands[cmd] = iresult
            if iresult:
                cvsFailures.append(realName)
                continue

            try:
                updatedf = open(realName, "w+")
                updatedf.write(resultStd)
                updatedf.close()
            except:
                # In case of failure, we remove the file
                if os.access(realName, os.F_OK):
                    os.unlink(realName)
                cvsFailures.append(realName)
            else:
                updateDict[obj_id] = statDict[obj_id]

        tmpStatus = self.filesToObjects(updateDict, createFolder=1)
        tmpStatus[1] = tmpStatus[1] + cvsFailures

        result = '<TABLE><TR><TD>\n' + \
                 '<PRE>%s</PRE>\n' %formatStatus(tmpStatus, "Imports") + \
                 '</TD></TR></TABLE>'

        os.chdir(oldDir)

        command = ''
        for i in listCommands.keys():
            command = command + '%s-> %d\n' %(i, listCommands[i])

        return MessageDialog(title='Check Out Returned',
                 message='Command:<PRE>%s</PRE><BR>\n%s\n' %(command,result),
                    action=REQUEST.form['FormPage'])

    def manage_objFromFS(self, obj_ids=None, isrec=1, REQUEST=None):
        """
        move objects from files to Zodb

        """

        # move to CVS directory...
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            return MessageDialog(title='Obj <- FS',
                       message='No object selected',
                       action=REQUEST.form['FormPage'])

        updateDict = {}

        for obj_id in obj_id_list:
            if statDict.has_key(obj_id) and statDict[obj_id].has_key('fname'):
                updateDict[obj_id] = statDict[obj_id]

        result = formatStatus(self.filesToObjects(updateDict), "Imports")
        result = "<table><tr><td><pre>%s</pre></td></tr></table>" % result

        os.chdir(oldDir)

        return MessageDialog(title='File to Object',
                             message='Update returned: %s' %  result,
                             action=REQUEST.form['FormPage'])


    def manage_checkIn(self, obj_ids=None, message="",
                       g_args='', cvs_args='', isrec=1, add_ids=None,
                       canCreate=0, REQUEST=None):
        """
        check in the current version(s)...
        """
        # TODO: canCreate is not used yet (id. for add_ids)

        # move to CVS directory...
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            return MessageDialog(title='Check In',
                     message='No object selected',
                     action=REQUEST.form['FormPage'])

        saveDict = {}
        missingObjects = []

        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if (ld['meta_type'] not in self.excludedTypes()):
                if ld.has_key('meta_type'):
                    saveDict[obj_id] = ld
                #else if canCreate and (not ld.has_key('ext')):
                #    add_id_list = self.checkList(add_ids)
                #    # Let's add this object

        # save to FS
        res = self.objectsToFiles(saveDict)
        res[1] = res[1] + missingObjects

        # convert to fnames
        fnames = self.idsToFileNames( saveDict, relNames=1 )

        if len(fnames) == 0:
            return MessageDialog(title='CheckIn...',
                     message='No object to check-in!',
                     action=REQUEST.form['FormPage'])

        if message == "":
            if self.allowEmptyCommit:
                message = "No Message"
            else:
                return MessageDialog(title='CheckIn...',
                         message='The commit message is empty!',
                         action=REQUEST.form['FormPage'])

        if self.useplainfiles:
            message = "'%s: %s'" %(REQUEST.AUTHENTICATED_USER, string.replace(message,"'",r"\'"))
        else:
            message = '%s: %s' %(REQUEST.AUTHENTICATED_USER, message)
            
        os.putenv('ZCVSUSERNAME', '%s' %REQUEST.AUTHENTICATED_USER)

        command = ['cvs'] + d_gargs + \
                  ['commit', '-m', message] + \
                  self.stringToList(cvs_args) + fnames

        resCVSval, resCVSstd, resCVSerr = self.processStreams(command)

        rv, statDict, ev = self.cvsStatusList(isrec)

        restoreDict = {}

        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if ld.has_key('meta_type') and \
               (ld['meta_type'] not in self.excludedTypes()):
                if ld.has_key('fnewerz') and ld['fnewerz']:
                    restoreDict[obj_id] = ld

        if resCVSval:
            result = '%s\n' %resCVSerr
        else:
            result = ''
        importStatus = self.filesToObjects(restoreDict)
        if (len(importStatus[0]) != 0) or (len(importStatus[1]) != 0):
            result = result + \
                     formatStatus(importStatus, "Exports")

        if result != '':
            result = "\n<table><tr><td>\n" + \
                     "<pre>%s</pre>\n" % result + \
                     "</td></tr></table>"

        os.chdir(oldDir)

        return MessageDialog(title='Commit Returned',
                     message='Command:<PRE>%s</PRE><BR>\nResult: %d<BR>%s'
                     % (srj(command), resCVSval, result),
                     action=REQUEST.form['FormPage'])

    def manage_objToFS(self, obj_ids=None, isrec=1, REQUEST=None):
        """
        check in the current version(s)...
        """

        # move to CVS directory...
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)

        if len(obj_id_list) == 0:
            return MessageDialog(title='FS <- Obj',
                     message='No object selected',
                     action=REQUEST.form['FormPage'])

        saveDict = {}
        missingObjects = []

        for obj_id in obj_id_list:
            ld = statDict[obj_id]
            if ld.has_key('meta_type'):
                if (ld['meta_type'] not in self.excludedTypes()):
                    saveDict[obj_id] = ld
            else:
                missingObjects.append(ld['location'] + ld['fname'])

        # save to FS
        tmpStatus = self.objectsToFiles(saveDict)
        os.chdir(oldDir)
        tmpStatus[1] = tmpStatus[1] + missingObjects

        return MessageDialog(title='Object Converted to Files',
                 message='Object Converted to Files:<BR>\n<PRE>%s</PRE>' \
                 %formatStatus(tmpStatus, "Exports"),
                 action=REQUEST.form['FormPage'])

    def manage_addfsdirCVS(self, g_args=''):
        """
        add dir to CVS
        """

        # move to CVS directory...
        oldDir = os.getcwd()

        if not os.access(self.getFSDir(), os.F_OK):
            os.mkdir(self.getFSDir())

        os.chdir(self.getFSDir())

        # add directory

        result = ''

        if not os.access(self.getFSDir() + '/CVS', os.F_OK):

            os.chdir(os.path.dirname(self.getFSDir()))
            bname = os.path.basename(self.getFSDir())

            d_gargs = self.stringToList(self.default_gargs)
            if (self.allow_gargs):
                d_gargs = d_gargs + self.stringToList(g_args);

            os.putenv('ZCVSUSERNAME', '')
            command = ['cvs'] + d_gargs + ['add', '%s' %bname]
            result = self.processCommands(command)
            command = ['cvs'] + d_gargs + \
                      ['commit', '-m', 'Add folder %s' %bname, '%s' %bname]
            result = result + self.processCommands(command)

            result = "<table><tr><td><pre>%s</pre></td></tr></table>" %(result)

        os.chdir(oldDir)

        return result

    def manage_addCVS(self, g_args='', obj_ids=None, add_ids=None,
                      message="No Message", cvs_args='',
                      isrec=1, REQUEST=None):
        """
        Add an object to the CVS...
        """

        # move to CVS directory...
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        rv, statDict, ev = self.cvsStatusList(isrec)

        obj_id_list = self.checkList(obj_ids)
        add_id_list = self.checkList(add_ids)

        saveDict = {}

        if len(obj_id_list) == 0:
            return MessageDialog(title='New object to CVS',
                       message='No object selected',
                       action=REQUEST.form['FormPage'])

        if message == "":
            if self.allowEmptyCommit:
                message = "No Message"
            else:
                return MessageDialog(title='New object to CVS',
                         message='The commit message is empty!',
                         action=REQUEST.form['FormPage'])

        resultOTF = []
        for obj_id in obj_id_list:
            try:
                saveDict[obj_id] = statDict[obj_id]
                saveDict[obj_id]['ext'] = saveDict[obj_id]['format_opt'] = \
                                      REQUEST.form['format;'+obj_id]
            except:
                resultOTF.append(obj_id)

        # save to FS
        succ,fail = self.objectsToFiles(saveDict)
        print "In addCVS succ=", succ, ": fail = ", fail

        for each_fail in fail:

            try:
                del saveDict[each_fail]
            except KeyError:
                pass
            
            resultOTF.append(each_fail)

        if saveDict:
            #
            # need to do two different adds.....
            # one for xml, another for zexp...
            #

            xmlDict = {}
            zexpDict = {}
            for k in saveDict.keys():
                if saveDict[k]['ext'] == 'xml':
                    xmlDict[k] = saveDict[k]
                else:
                    zexpDict[k] = saveDict[k]

            os.putenv('ZCVSUSERNAME', '%s' %REQUEST.AUTHENTICATED_USER)

            if message=='':
                message = 'New file'
            message = '%s: %s' %(REQUEST.AUTHENTICATED_USER, message)

            # xml
            if len(xmlDict.keys())>0:
                addList = self.idsToFileNames( xmlDict, relNames=1)
                command = ['cvs'] + d_gargs + ['add'] + \
                          self.stringToList(cvs_args) + addList
                # do the update
                result1, resultStd1, resultErr1 = self.processStreams(command)

                # we commit now, because parseStatusList fails if a FS object
                # isn't found in the CVS tree...
                command = ['cvs'] + d_gargs + \
                          ['commit', '-m', message] + \
                          self.stringToList(cvs_args) + addList
                rescom1, rescomStd1, rescomErr1 = self.processStreams(command)
                result1 = "xml:(add->%d, commit->%d)\n\n" %(result1, rescom1)
            else:
                result1 = ''

            # zexp
            if len(zexpDict.keys())>0:
                addList = self.idsToFileNames(zexpDict, relNames=1)
                command = ['cvs'] + d_gargs + ['add', '-kb'] + \
                          self.stringToList(cvs_args) + addList
                # do the update
                result2 = self.processCommands( command )

                # we commit now, because parseStatusList fails if a FS object
                # isn't found in the CVS tree...
                command = ['cvs'] + d_gargs + \
                          ['commit', '-m', message] + \
                          self.stringToList(cvs_args) + addList
                rescom2, rescomStd2, rescomErr2 = self.processStreams(command)
                result2 = "zexp:(add->%d, commit->%d)\n\n" %(result2, rescom2)
            else:
                result2 = ''

            if len(resultOTF) != 0:
                result = '<BR>Failures: %s<BR>' %resultOTF
            else:
                result = ''

            result = result + \
                     "<table><tr><td><pre>%s</pre></td></tr></table>" \
                     % (result1 + "\n\n" + result2)
        else:
            result = "No Objects to add..."
            if len(resultOTF) != 0:
                result = result + '<BR>Failures: %s<BR>' %resultOTF

        os.chdir(oldDir)
        return MessageDialog(title='Add Returned',
                             message='Add returned:<BR>\n%s' %  result,
                             action=REQUEST.form['FormPage'])

    def manage_cvsAdmin(self, cvs_admin_args='', g_args='', adminCommand='',
                        obj_ids=None, isrec=1, REQUEST=None):
        """
        perform some cvs administration
        """

        if REQUEST is None:
            REQUEST = {}

        #import pdb
        #pdb.set_trace()
        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'We are in cvsAdmin')

        obj_id_list = self.checkList(obj_ids)
        rv, statDict, ev = self.cvsStatusList(isrec)

        if len(obj_id_list) == 0:
            return MessageDialog(title='Admin',
                       message='No object selected',
                       action=REQUEST.form['FormPage'])

        adminDict = {}
        tmpStatus = ''

        for obj_id in obj_id_list:
            adminDict[obj_id] = statDict[obj_id]

        # convert to fnames
        fnames = self.idsToFileNames( adminDict, relNames=1 )

        if adminCommand == 'removeFS':
            for name in fnames:
                os.unlink(name)
            command = 'Files removed'
        elif adminCommand == 'removeZ':
            for itemId in adminDict.keys():
                self._containerDelObject( itemId )
            result = "Deleted IDs: %s" % `adminDict.keys()`
            command = ''
        else:
            # if 'diff': convert the Zope object into file
            if adminCommand == 'diff':
                saveDict = {}
                for obj_id in obj_id_list:
                    ld = statDict[obj_id]
                    if ld.has_key('meta_type') and \
                       (ld['meta_type'] not in self.excludedTypes()):
                        saveDict[obj_id] = ld
                tmpStatus = self.objectsToFiles(saveDict)
                if (len(tmpStatus[1])):
                    tmpStatus = 'Error in objectToFiles<BR>\n'
                else:
                    tmpStatus = ''

            command = ['cvs'] + d_gargs + ['-l', adminCommand] + \
                      self.stringToList(cvs_admin_args) + fnames
            # do the update
            result, resultStd, resultErr = self.processStreams(command)

            try:
                resultStd = html_quote.html_quote(resultStd)
            except TypeError:
                # backward compat when html_quote was a function, not a module
                resultStd = html_quote(resultStd)

            if adminCommand == 'diff':
                resultStd = self.colorizeDiff(resultStd, command)
            else:
                resultStd = "<pre>" + resultStd + "</pre>"

            result = tmpStatus + \
                         "<table><tr><td>%s</td></tr></table>" \
                         % resultStd

        os.chdir(oldDir)

        if command:
            return MessageDialog(title='Admin Returned',
                     message='Command:<PRE>%s</PRE><BR>Result:<BR>%s'
                     % (srj(command), result),
                     action=REQUEST.form['FormPage'])
        else:
            return MessageDialog(title='Admin Returned',
                     message='%s<br>Returned: %d'
                     % (srj(command), result),
                     action=REQUEST.form['FormPage'])


    def removeObject(self, obj):
        if string.find(obj, '/') != -1:
            # Recursive call
            folder, newobj = string.split(obj, '/', 1)
            try:
                thisObj = self._containerGetattr(folder)
            except:
                return 1  # Failure
            if hasattr(thisObj, 'isAnObjectManager') and \
               thisObj.isAnObjectManager:
                return thisObj.CVS.getTimeTracker(newobj_id)
            else:
                raise 'fileToObject', "%s is not an object manager" %folder
        try:
            self._containerDelObject(obj)
        except:
            return 1  # Failure
        return 0  # Ok

    def manage_remove(self, obj_ids=None, message="No Message",
                      g_args='', cvs_args='', what=0, isrec=1, REQUEST=None):
        """
        remove:
        - what == 0: only FS object(s)
        - what == 1: Zope, FS and CVS objects
        """

        if REQUEST is None:
            REQUEST = {}

        oldDir = os.getcwd()
        os.chdir(self.getFSDir())

        d_gargs = self.stringToList(self.default_gargs)
        if (self.allow_gargs):
            d_gargs = d_gargs + self.stringToList(g_args);

        zLOG.LOG('ZCVSMixin', zLOG.BLATHER, 'We are in manage_remove')

        obj_id_list = self.checkList(obj_ids)
        rv, statDict, ev = self.cvsStatusList(isrec)

        adminDict = {}
        command = command2 = ''
        result = result2 = ''

        for obj_id in obj_id_list:
            if statDict.has_key(obj_id):
                adminDict[obj_id] = statDict[obj_id]

        if len(adminDict) == 0:
            return MessageDialog(title='Remove...',
                       message='No object to remove!',
                       action=REQUEST.form['FormPage'])

        # convert to fnames
        fnames = self.idsToFileNames(adminDict, relNames=1)

        if (what == 1):
            # Let's remove everything
            delIds = []
            for itemId in adminDict.keys():
                if (self.removeObject(itemId) == 0):
                    delIds.append(itemId)
            result_obj = "Deleted IDs: %s" %delIds

            command = ['cvs'] + d_gargs + ['remove', '-f'] + \
                      self.stringToList(cvs_args) + fnames
            result, resStd, resErr = self.processStreams(command)

            msg = srj(result_obj, ', ') + \
                  '<BR>Command:<PRE>%s</PRE><BR>Result:<BR>%s' \
                  % (srj(command), result)
            if result:
                msg = msg + \
                      '<BR>Error:<BR>\n<TABLE><TR><TD>\n' + \
                      '<PRE>%s</PRE>\n' %resErr + \
                      '</TD></TR></TABLE>'

            try:
                message = '%s: %s' %(REQUEST.AUTHENTICATED_USER, message)
            except:
                os.putenv('ZCVSUSERNAME', '')
            else:
                os.putenv('ZCVSUSERNAME', '%s' %REQUEST.AUTHENTICATED_USER)

            command2 = ['cvs'] + d_gargs + ['commit', '-m', message] + \
                       self.stringToList(cvs_args) + fnames
            result2, resStd, resErr = self.processStreams(command2)

            msg = msg + '<HR>Command:<PRE>%s</PRE><BR>Result:<BR>%s' \
                  % (srj(command2), result2)
            if result2:
                msg = msg + \
                      '<BR>Error:<BR>\n<TABLE><TR><TD>\n' + \
                      '<PRE>%s</PRE>\n' %resErr + \
                      '</TD></TR></TABLE>'
        else:
            # Let's remove fnames
            for name in fnames:
                if os.access(name, os.F_OK):
                    os.unlink(name)
            msg = '%s deleted<BR>' %fnames

        os.chdir(oldDir)

        if (what == 1):
            result = MessageDialog(title='Admin Returned',
                         message=msg,
                         action='cvsStatusForm')
        else:
            result = MessageDialog(title='Admin Returned',
                         message=msg, action=REQUEST.form['FormPage'])
        return result

    def manage_removeAll(self, obj_ids=None, message="No Message",
                         g_args='', cvs_args='', what=1, isrec=1, REQUEST=None):
        """
        remove Zope, FS and CVS objects
        """
        return self.manage_remove(obj_ids, message, g_args, cvs_args,
                                  what=1, isrec=isrec, REQUEST=REQUEST)

    def fileToObject(self, obj_id, obj, createFolder=0):
        """
        import one file into this container, or into a child Zope folder...
        (this is a recursive function)
        if the folder doesn't exist and createFolder is true, then the folder
        is created
        """
        if string.find(obj_id, '/') != -1:
            # Recursive call
            folder, newobj_id = string.split(obj_id, '/', 1)
            if self._containerHasID(folder):
                thisObj = self._containerGetattr(folder)
            else:
                if createFolder and folder != "acl_users":
                    # "acl_users" is a special case: most likely it is an
                    # exUserFolder that contains regular zope objects and an
                    # CVS Folder as a child.  We do not want to create a plain
                    # Folder in place of it during a fresh checkout.

                    # We create "folder", and we add a ZCVS object into it.
                    self.manage_addFolder(folder)
                    # We commit the transaction, so that we can import files
                    # into "folder"
                    get_transaction().commit()
                    # Begin another transaction
                    get_transaction().begin()
                    self.branchToCVS(1)
                    thisObj = self._containerGetattr(folder)
                else:
                    return (1, string.split(obj_id, '/')[-1])
            if hasattr(thisObj, 'isAnObjectManager') and thisObj.isAnObjectManager:
                return thisObj.CVS.fileToObject(newobj_id, obj, createFolder)
            else:
                raise 'fileToObject', "%s is not an object manager" %folder

        # Let's import the file into *this* container
        if obj.has_key('ext'):
            ext = obj['ext']
            fname = self.idToFileName(obj_id, ext)
            try:
                try:
                    ob=self._container_p_jar().importFile(fname,
                       customImporters=customImporters)
                except:
                    ob=self._container_p_jar().importFile(fname)

                id=ob.id
                if hasattr(id, 'im_func'):
                    id=id()

                if self._containerHasID(id):
                    self._containerDelObject(id)

                self._containerSetObject(id, ob)

                btime=self._containerGetattr(id).bobobase_modification_time

                if hasattr(btime, 'im_func'):
                    btime = btime()

                btime = btime.timeTime()

                tracker = self.getTimeTracker(id)
                tracker.bmtime = btime
                tracker.fsmtime = os.path.getmtime(fname)
                self.setTimeTracker(obj['location'] + obj['zname'], tracker)
                return (0, fname)  # Success
            except:
                f = StringIO.StringIO()
                traceback.print_exc(file=f)
                f.seek(0)
                return (1, fname)  # Failure
        return (1, fname)  # Failure

    def filesToObjects(self, obj_statDict, createFolder=0):
        """
        import set of files from dict into this container or a 
        child container...
        """
        failedImports = []
        successfulImports = []

        #do_break()

        for k in obj_statDict.keys():
            state, name = self.fileToObject(k, obj_statDict[k], createFolder)
            if state:
                failedImports.append(name)
            else:
                successfulImports.append(name)

        return [successfulImports, failedImports]

    def objectsToFiles(self, obj_statDict):
        """
        move objects from statDict from this Zope container to file system,
          return list of filenames... set modification time of
          object to be the same as the bobobase modification time
          if the object has been modified, otherwise set it to
          the old fstime. Make sure there is an extension for each file...
          otherwise it will default to zexp....
        """

        self.checkTracker()

        failedImports = []
        successfulImports = []

        obj_id_list = obj_statDict.keys()

        for obj_id in obj_id_list:
            currItem = obj_statDict[obj_id]
            ob=currItem['zobj']

            if (currItem.has_key('ext') and currItem['ext'] == 'xml') or \
               (currItem.has_key('format_opt')
               and currItem['format_opt'] == 'xml'):
                fname = self.idToFileName(obj_id, 'xml')
                if ob is not None:
                    if self.cleanupXML:
                        doExport(ob._p_jar, ob._p_oid, fname)
                    else:
                        from OFS import XMLExportImport
                        XMLExportImport.exportXML(ob._p_jar, ob._p_oid, fname)
                else:
                    if os.access(fname, os.F_OK):
                        os.unlink(fname)
                    failedImports.append(fname)
                    continue
            else:
                fname = self.idToFileName(obj_id, 'zexp')
                if ob is not None:
                    ob._p_jar.exportFile(ob._p_oid, fname)
                else:
                    if os.access(fname, os.F_OK):
                        os.unlink(fname)
                    failedImports.append(fname)
                    continue

            # keep mtime same as btime...
            if ob is not None:
                btime = ob.bobobase_modification_time
            else:
                btime = 0.0
            if hasattr(btime, 'im_func'):
                btime = btime().timeTime()

            tracker = self.getTimeTracker(obj_id)

            if tracker.bmtime < btime:
                # this one has *really* changed since import
                tracker.bmtime = btime
                # update trackerDict...
                self.setTimeTracker(obj_id, tracker)
                fstime = btime
            else:
                # not a real change.. just import itself
                fstime = tracker.fsmtime


            if os.access(fname, os.F_OK):
                os.utime(fname, (fstime, fstime))
                tracker.fsmtime = fstime
                self.setTimeTracker(obj_id, tracker)

            successfulImports.append(fname)

        return [successfulImports, failedImports]

    def getTrackerDict(self):

        """
        Get the tracker dict....

        """
        self.checkTracker()

        return self.trackerDict

    def formatResult(self, command, resultVal, resultStd, resultErr):

        result = ["Command: '%s' returns %i" % (srj(command), resultVal)]
        result.append("StdOut:\n%s" % resultStd)
        result.append("StdErr:\n%s" % resultErr)

        return string.join(result,'\n')

    def stringToList(self, s):
        l = []
        c = 0;
        eop = 0
        char = ''
        while (c < len(s)):
            if ((ord(s[c]) == 32) and (char == '')):
                if c:
                    s2 = s[0:c]
                    if ((s2[0] == s2[-1]) and (s2[0] in ['\'', '"'])):
                        l = l + [s2[1:-1]]
                    else:
                        l = l + [s2]

                s = s[c+1:]
                c = 0
            else:

                if (char != ''):
                    if (s[c] == char):
                        char = ''
                else:
                    if ((s[c] == '\'') or (s[c] == '"')):
                        char = s[c]
                c = c + 1

        if (len(s)):
            if ((s[0] == s[-1]) and (s[0] in ['\'', '"'])):
                l = l + [s[1:-1]]
            else:
                l = l + [s]
        return l

    def colorizeDiff(self, resultStr, command):
        "Colorize cvs diff output"

        import re

        colorize = lambda style, text=r"\1": '<div class="%s">%s</div>' % (style, text);
        color_fg = lambda style, text=r"\1": '<span class="%s">%s</span>' % (style, text);

        stylesheet = """
<style type="text/css">
<!--
  div.file      { color: #000000; background-color: #dddddd }
  div.oldfile   { color: #6050cc; background-color: #dddddd }
  div.newfile   { color: #108028; background-color: #dddddd }
  div.chunk     { color: #aa2222; background-color: #eeeeee }
  div.oldnoise  { color: #000000; background-color: #fef8ff }
  div.newnoise  { color: #000000; background-color: #f8fffe }
  div.old       { color: #000000; background-color: #e3e0ff }
  div.new       { color: #000000; background-color: #e0ffe6 }
  span.tag      { color: #000080 }
  span.string   { color: #aa0000 }
  span.entity   { color: #b75500 }
  span.newline  { color: #888888 }
  -->
</style>
"""

        # diff output
        resultStr = re.sub('<pre>Index:', '<pre>\nIndex:', resultStr)
        resultStr = re.sub(r'(?m)^((?:Index:|==|RCS|retrieving|diff).*)$',      colorize("file"), resultStr)

        # noise
        noise = r'&lt;(?:record|tuple|string|dictionary|object|global|list|reference) id='

        if '-u' in command:
            # unified diff
            resultStr = re.sub(r'(?m)^(---.*)$',    colorize("oldfile"), resultStr)
            resultStr = re.sub(r'(?m)^(\+\+\+.*)$', colorize("newfile"), resultStr)
            resultStr = re.sub(r'(?m)^(@@.*)$',     colorize("chunk"), resultStr)
            resultStr = re.sub(r'(?m)^(-.*%s.*)$' % noise,  colorize("oldnoise"), resultStr)
            resultStr = re.sub(r'(?m)^(\+.*%s.*)$' % noise, colorize("newnoise"), resultStr)
            resultStr = re.sub(r'(?m)^(-.*)$',      colorize("old"), resultStr)
            resultStr = re.sub(r'(?m)^(\+.*)$',     colorize("new"), resultStr)
        elif '-y' not in command and '-e' not in command and '-f' not in command:
            # context diff
            resultStr = re.sub(r'(?m)^(&lt;.*%s.*)$' % noise, colorize("oldnoise"), resultStr)
            resultStr = re.sub(r'(?m)^(&gt;.*%s.*)$' % noise, colorize("newnoise"), resultStr)
            resultStr = re.sub(r'(?m)^(&lt;.*)$',   colorize("old"), resultStr)
            resultStr = re.sub(r'(?m)^(&gt;.*)$',   colorize("new"), resultStr)

        # XML syntax
        resultStr = re.sub(r'(&lt;)(.*?)(&gt;)',    r"\1" + color_fg("tag", r"\2") + r"\3", resultStr)
        resultStr = re.sub(r'(&quot;.*?&quot;)',    color_fg("string"), resultStr)
        resultStr = re.sub(r'(&amp;(?:[a-z_-]+|#[0-9xa-fA-F]+);)', color_fg("entity"), resultStr)
        resultStr = re.sub(r'(\\[rn])',             color_fg("newline"), resultStr)

        # Postprocess for Mozilla
        resultStr = re.sub('</div>\n', '\n</div>', resultStr)

        return stylesheet + "<pre>" + resultStr + "</pre>"

    #def DBGwrite(self, str):
    #    dfdbg = open("/tmp/zopeCVS.log", "a+")
    #    dfdbg.write(str)
    #    dfdbg.close()


def quoteArgs( aList ):
    return string.join(map(lambda x: "'%s'" % x, aList), ' ')

def splitResizeJoin( aString, maxSize = 70 ):
    newLines = []

    if type(aString) is types.ListType:
        aString = string.join(aString)

    words = string.split(aString)
    currString = ''

    for word in words:
        if len(word) + len(currString) > maxSize:
            newLines.append(currString)
            currString = ''
        currString = currString + ' ' + word
    if currString:
        newLines.append(currString)

    return string.join(newLines,'\n')

srj = splitResizeJoin

def formatStatus(status, what, all=0):
    """
    status[0] == successful transactions
    status[1] == failed transactions
    what: 'Imports', 'Exports'...
    if all == 0, only non-empty lists are shown
    """
    res = ""
    if all or (len(status[0]) != 0):
        res = res + "<b>Successful %s:</b>\n" %what + \
                    string.join(status[0],',\n') + "<br>\n"
    if all or (len(status[1]) != 0):
        res = res + "<b>Failed %s:</b>\n" %what + \
                    string.join(status[1],',\n') + "<br>\n"
    return res

def test():
    x = ZCVSMixin()
    x._v_debug = 1
    x.fsroot = '/tmp/zopeObjs/'
    x.fsrepos = 'testCVS'
    x.active = 1
    x.useplainfiles = 1

    print x.getFSRoot()

    import pdb
    pdb.set_trace()
    
    a, theDict, b = x.cvsStatusList(1)

    print a, b
    for k in theDict.keys():
        print `k`
        for j in theDict[k].keys():
            print '     ',j,':', `theDict[k][j]`


if __name__=="__main__":
    test()
